module Main where

import System.Environment
import Control.Parallel.Eden

import ParMap(unshuffle, parMapEden, parmapReduceStream)
import DC

usage = "Arguments: version(seq,dc,mapRed) and problem size"

main :: IO ()
main = do args <- getArgs
       	  if length args < 2
	    then putStrLn usage
	    else let [v,n] = args 
                 in putStrLn ((rnf (ms v ((read n)::Int))) 
                              `seq` "Done")

ms :: String -> Int -> [Int]
ms "seq" n = mergeSort (list n)
ms "dc"  n = unBox $ mergeSort_skel  (Box (list n))
ms "mapRed" n 
   = let nodes = noPe-1
         reducer (Box l1) (Box l2) = Box (sortMerge l1 l2)
         mapper (Box xs) = Box (mergeSort xs)
         input = map Box (unshuffle nodes (list n))::[Box [Int]]
     in unBox (parmapReduceStream nodes mapper reducer (Box []) input)

newtype Box a = Box {unBox :: a} deriving Show
instance Trans a => Trans (Box a)
instance NFData a => NFData (Box a)
  where rnf (Box a) = rnf a
  
sortMerge :: Ord a => [a] -> [a] -> [a]
sortMerge []           ylist        = ylist
sortMerge xlist        []           = xlist
sortMerge xlist@(x:xs) ylist@(y:ys)
   | x <= y = x : sortMerge xs ylist
   | x >  y = y : sortMerge xlist ys


list :: Int -> [Int]
list n = [n, n-1..1]  

-- using skeleton
mergeSort_skel      :: (Ord a, Trans a) => Box [a] -> Box [a]
mergeSort_skel = dc_c (2* dlog noPe) triv solve split combine
  where triv (Box xs) = null xs || null (tail xs)
        solve = id 
        split (Box xs) = map Box (unshuffle 2 xs)
        combine [Box rs1, Box rs2] = Box (sortMerge rs1 rs2)

--version without skeleton
mergeSort      :: (Ord a, Trans a) => [a] -> [a]
mergeSort []   =  []
mergeSort [x]  =  [x]
mergeSort xs   =  sortMerge (mergeSort xs1)
                            (mergeSort xs2)
  where [xs1,xs2] = unshuffle 2 xs

dlog :: Int -> Int
dlog n = dlog' 1 n
    where dlog' m n | n<=1 = m
                    | otherwise = dlog' (m+1) (n `div` 2)
